import sys

import torch.optim

import runai.utils

# this design is inspired by Horovod's `DistributedOptimizer`
def Optimizer(optimizer, steps):
    """
    Wraps any valid PyTorch optimizer with gradient accumulation
    """
    class _GradientAccumulationOptimizer(torch.optim.Optimizer):
        class State:
            def __init__(self, steps):
                self.steps = steps
                self.step = 0

        def __init__(self, steps, params):
            self.__class__.__optimizer__.__init__(self, params) # TODO(levosos): what about `defaults`?
            self.state['runai'] = _GradientAccumulationOptimizer.State(steps)

            runai.utils.log.debug('Wrapping \'%s\' PyTorch optimizer with GA of %d steps', optimizer.__class__.__name__, steps)

        def step(self, closure=None):
            loss = None

            if self.runai.step % self.runai.steps == self.runai.steps - 1:
                loss = self.__class__.__optimizer__.step(self, closure)
            else:
                if closure is not None:
                    with torch.enable_grad():
                        loss = closure() # TODO(levosos): is this good?

            self.runai.step += 1
            return loss

        def zero_grad(self):
            if self.runai.step % self.runai.steps == 0:
                self.__class__.__optimizer__.zero_grad(self)

        @property
        def runai(self):
            return self.state['runai']

    # the main idea is to dynamically create a class that has all the functionality of the passed optimizer
    # (this is done by inheriting it) while overriding `step()` and `zero_grad()` to accumulate the gradients
    # and actually assign and zero them once in a few steps
    d = dict(_GradientAccumulationOptimizer.__dict__)
    d['__optimizer__'] = optimizer.__class__

    cls = type(
        optimizer.__class__.__name__,
        (optimizer.__class__,),
        d
    )
    return cls(steps, optimizer.param_groups)

# declare a GA version of builtin optimizers
def _optimizer(optimizer):
    setattr(
        sys.modules[__name__],
        optimizer,
        lambda steps, *args, **kwargs: Optimizer(getattr(torch.optim, optimizer)(*args, **kwargs), steps)
    )

[_optimizer(optimizer) for optimizer in ['Adadelta', 'Adagrad', 'Adam', 'AdamW', 'Adamax', 'ASGD', 'SGD', 'Rprop', 'RMSprop']]



import time

import keras

from runai.reporter import reportMetric, reportParameter

__original_fit__ = keras.Model.fit
__original_compile__ = keras.Model.compile
__original_fit_generator__ = keras.Model.fit_generator

def disableAutoLog():
    keras.Model.fit = __original_fit__
    keras.Model.fit_generator = __original_fit_generator__
    keras.Model.compile = __original_compile__

def autolog(accuracy=True, loss=True, learning_rate=True, epoch=True, step=True, batch_size=True, overall_epochs=True,
            optimizer_name=True, number_of_layers=True, loss_method=False, epsilon=False):
    # The following line must be at top of the method
    autolog_inputs = locals()

    def fit(self, *args, **kwargs):
        _append_autolog_metrics_to_callbacks(args, kwargs, index_in_args=5)
        _report_overall_epochs(args, kwargs, index_in_args=3)
        _report_batch_size(args, kwargs)

        return __original_fit__(self, *args, **kwargs)

    def fit_generator(self, *args, **kwargs):
        _append_autolog_metrics_to_callbacks(args, kwargs, index_in_args=4)
        _report_overall_epochs(args, kwargs, index_in_args=2)

        return __original_fit_generator__(self, *args, **kwargs)

    def compile(self, *args, **kwargs):
        _add_autolog_metrics(args, kwargs)
        _report_loss_method(args, kwargs)
        _report_loss_method(args, kwargs)

        return __original_compile__(self, *args, **kwargs)

    keras.Model.fit = fit
    keras.Model.fit_generator = fit_generator
    keras.Model.compile = compile


    def _report_overall_epochs(original_args, original_kwargs, index_in_args=None):
        if index_in_args is None:
            raise ValueError("'index_in_args' must be specified")

        if _should_report_metric_or_parameter(autolog_inputs, 'overall_epochs'):
            overall_epochs_val = _get_value_of_method_parameter(original_args, original_kwargs, index_in_args, key_in_kwargs='epochs', default_value=1)  # Keras' default value
            reportMetric('overall_epochs', overall_epochs_val)

    def _report_batch_size(original_args, original_kwargs):
        if _should_report_metric_or_parameter(autolog_inputs, 'batch_size'):
            batch_size_val = _get_value_of_method_parameter(original_args, original_kwargs, index_in_args=2, key_in_kwargs='batch_size')
            if batch_size_val:
                reportMetric('batch_size', batch_size_val)

    def _report_loss_method(original_args, original_kwargs):
        loss_method_val = _get_value_of_method_parameter(original_args, original_kwargs, index_in_args=1, key_in_kwargs='loss')
        if loss_method_val:
            _report_parameter_if_needed(autolog_inputs, 'loss_method', loss_method_val)

    def _append_autolog_metrics_to_callbacks(original_args, original_kwargs, index_in_args=None):
        if index_in_args is None:
            raise ValueError("'index_in_args' must be specified")

        if 'callbacks' in original_kwargs:
            original_kwargs['callbacks'].append(KerasAutoMetricReporter())
        elif index_in_args < len(original_args):
            original_args[index_in_args].append(KerasAutoMetricReporter())
        else:
            original_kwargs['callbacks'] = [KerasAutoMetricReporter()]

    def _get_value_of_method_parameter(original_args, original_kwargs, index_in_args=None, key_in_kwargs=None, default_value=None):
        if index_in_args is None:
            raise ValueError("'index_in_args' must be specified")

        if key_in_kwargs is None:
            raise ValueError("'key_in_kwargs' must be specified")

        if key_in_kwargs in original_kwargs:
            return original_kwargs[key_in_kwargs]
        elif index_in_args < len(original_args):
            return original_args[index_in_args]
        return default_value

    def _add_autolog_metrics(original_args, original_kwargs):
        metrics_index_in_args = 2
        if len(original_args) > metrics_index_in_args:
            metrics = original_args[metrics_index_in_args]
        else:
            if 'metrics' not in original_kwargs:
                original_kwargs['metrics'] = []

            metrics = original_kwargs['metrics']

        autolog_metrics_from_logs = ['acc']
        for metric in autolog_metrics_from_logs:
            if metric not in metrics:
                metrics.append(metric)

    def _report_metric_if_needed(autolog_inputs, key, value):
        if _should_report_metric_or_parameter(autolog_inputs, key):
            reportMetric(key, value)

    def _report_parameter_if_needed(autolog_inputs, key, value):
        if _should_report_metric_or_parameter(autolog_inputs, key):
            reportParameter(key, value)

    def _should_report_metric_or_parameter(autolog_inputs, key):
        return key in autolog_inputs and autolog_inputs[key]

    class KerasAutoMetricReporter(keras.callbacks.Callback):
        def on_train_begin(self, logs=None):
            _report_parameter_if_needed(autolog_inputs, 'optimizer_name', type(self.model.optimizer).__name__)
            _report_metric_if_needed(autolog_inputs, 'number_of_layers', len(self.model.layers))

            self._report_parameter_from_model_optimizer('learning_rate', 'lr')
            self._report_parameter_from_model_optimizer('epsilon')

        def on_batch_end(self, batch, logs={}):
            _report_metric_if_needed(autolog_inputs, 'step', batch)
            self._report_metric_from_logs_if_needed(autolog_inputs, "acc", logs, metric_name="accuracy")
            self._report_metric_from_logs_if_needed(autolog_inputs, "loss", logs)

        def on_epoch_begin(self, epoch_val, logs=None):
            _report_metric_if_needed(autolog_inputs, 'epoch', epoch_val)

        def on_epoch_end(self, epoch_val, logs=None):
            _report_metric_if_needed(autolog_inputs, 'epoch', epoch_val)
            self._report_parameter_from_model_optimizer('learning_rate', 'lr')

        def _report_parameter_from_model_optimizer(self, metric_name, name_of_optimizer_attr=None):
            if not name_of_optimizer_attr:
                name_of_optimizer_attr = metric_name

            if not _should_report_metric_or_parameter(autolog_inputs, metric_name) or not hasattr(self.model.optimizer, name_of_optimizer_attr):
                return

            parameter_from_model = getattr(self.model.optimizer, name_of_optimizer_attr)
            value = parameter_from_model if type(parameter_from_model) is float else keras.backend.eval(parameter_from_model)
            reportParameter(metric_name, value)

        def _report_metric_from_logs_if_needed(self, autolog_inputs, key_in_logs, logs, metric_name=None):
            if key_in_logs not in logs:
                return

            if not metric_name:
                metric_name = key_in_logs

            _report_metric_if_needed(autolog_inputs, metric_name, logs[key_in_logs])



import enum
from os import environ

from prometheus_client import CollectorRegistry, Gauge, pushadd_to_gateway

GROUPING_KEY = "podUUID"
GATEWAY_URL_KEY = "reporterGatewayURL"
PUSH_GATEWAY_JOB_NAME = "reporter_pod_info"
REPORTER_PUSH_GATEWAY_METRIC_PREFIX = "reporter_push_gateway_metric"
REPORTER_PUSH_GATEWAY_METRIC_PARAMETER = "reporter_push_gateway_parameter"

class ReportType(enum.Enum):
    metric = 1
    parameter = 2

def reportMetric(reporter_metric_name, reporter_metric_value):
    createGaugeAndPushToGateway(reporter_metric_name, reporter_metric_value, ReportType.metric)

def reportParameter(reporter_param_name, reporter_param_value):
    createGaugeAndPushToGateway(reporter_param_name, reporter_param_value, ReportType.parameter)

def createGaugeAndPushToGateway(reporter_name, reporter_value, report_type):
    registry = CollectorRegistry()

    if report_type is ReportType.metric:
        label_names = ['metric_name', 'push_gateway_type']
        label_values = [reporter_name, 'metric']
        gauge_name = REPORTER_PUSH_GATEWAY_METRIC_PREFIX + "_" + reporter_name
        gauge_value = reporter_value

    else:
        label_names = ['param_name', 'param_value', 'push_gateway_type']
        label_values = [reporter_name, reporter_value, 'parameter']
        gauge_name = REPORTER_PUSH_GATEWAY_METRIC_PARAMETER + "_" + reporter_name
        gauge_value = 1

    gauge = Gauge(name=gauge_name, documentation="",labelnames=label_names, registry=registry)

    gauge.labels(*label_values).set(gauge_value)

    pushadd_to_gateway(gateway=environ[GATEWAY_URL_KEY], job=PUSH_GATEWAY_JOB_NAME,
                       registry=registry, grouping_key={GROUPING_KEY: environ[GROUPING_KEY]})



import keras.layers
import tensorflow as tf

import runai.utils

from . import coordinator

class Keep(keras.layers.Layer):
    """ A MP-supported generic Keras layer

    The main goal of this class is to ease the implementation of MP-supported
    Keras layers in order to avoid merging an already parallelised layer

    In case that the input tensor is a Run:AI parallelised layer, we duplicate
    this layer and activate it on every part of the parallelised input layer.
    Otherwise, the layer is activated on the non-parallelised input layer,
    just as it was supposed to run originally, and the input tensor is not
    being split. Anyways, the layer actions and algorithm are kept untouched.

    If the input layer is indeed a Run:AI parallelised one, the merged tensor
    is the concatenation of the layer's outputs. The channel dimension is taken
    from the 'data_format' attribute, if one exists, otherwise it is assumed to
    be the last dimension.

    The following example shows the essence of the inheritance being done, where:
        1) 'Layer' is keras.layers.Layer
        2) 'Keras' is the Keras implementation of a layer (e.g. keras.layers.Dropout)
        3) 'Runai' is this very class (runai.mp.layers.Keep)
        4) 'Final' is the MP-supported Keras layer implementation

    class Layer(object):
        def call(self):
            print('Layer.call()')

    class Keras(Layer):
        def call(self):
            print('Keras.call()')
            super(Keras, self).call()

    class Runai(Layer):
        def call(self):
            print('Runai.call()')
            super(Runai, self).call()

    class Final(Runai, Keras): pass
    """
    def call(self, inputs):
        if isinstance(inputs, (tuple, list)):
            valid = all([coordinator.registered(input) for input in inputs])
        else:
            valid = coordinator.registered(inputs)

        if not valid:
            runai.utils.log.debug('Not parallelising \'%s\' layer "%s"', self.__class__.__name__, getattr(self, 'name', 'N/A'))
            return super(Keep, self).call(inputs)

        runai.utils.log.info('Using parallelised input for \'%s\' layer "%s"', self.__class__.__name__, getattr(self, 'name', 'N/A'))

        channel_axis = 1 if getattr(self, 'data_format', 'channels_last') == 'channels_first' else -1

        def output(gpu, inputs):
            with tf.device('/device:GPU:%d' % gpu):
                return super(Keep, self).call(inputs)

        if isinstance(inputs, (tuple, list)):
            inputs = [coordinator.resolve(input) for input in inputs]
            inputs = [[input[i] for input in inputs] for i in range(len(inputs[0]))]
        else:
            inputs = coordinator.resolve(inputs)
        
        outputs = [output(gpu, input) for gpu, input in enumerate(inputs)]
        merged = keras.layers.Concatenate(axis=channel_axis)(outputs)
        coordinator.register(merged, outputs)
        return merged

    @staticmethod
    def create(name):
        """ Create an MP-supported Keras layer with 'keep' logic
        """
        return type(name, (Keep, getattr(keras.layers, name)), {})



import sys

from .keep import Keep

def _layer(layer):
    setattr(
        sys.modules[__name__],
        layer,
        Keep.create(layer)
    )

[_layer(layer) for layer in [
    'Add',
    'Subtract',
    'Multiply',
    'Average',
    'Maximum',
    'Minimum',
]]

def _method(method):
    setattr(
        sys.modules[__name__],
        method,
        lambda inputs, *args, **kwargs: getattr(sys.modules[__name__], method.capitalize())(*args, **kwargs)(inputs) # mimicking the method implementations at keras/layers/merge.py
    )

[_method(method) for method in [
    'add',
    'subtract',
    'multiply',
    'average',
    'maximum',
    'minimum'
]]



import sys

import keras.optimizers
import keras.backend as K

import runai.utils

from . import hooks

class Optimizer(keras.optimizers.Optimizer):
    def __init__(self, optimizer, steps):
        super(Optimizer, self).__init__()
        self.optimizer = optimizer
        self.steps = steps

        runai.utils.log.debug('Wrapping \'%s\' Keras optimizer with GA of %d steps', optimizer.__class__.__name__, steps)

    def get_updates(self, loss, params):
        grads = self.get_gradients(loss, params)

        with K.name_scope("runai") as name_scope:
            iterations = K.variable(0, dtype='int64', name='iterations')
            first      = K.equal(iterations % self.steps, 0)
            last       = K.equal(iterations % self.steps, self.steps - 1)

            # variables to hold the accumulated gradients between steps
            vagrads = [K.zeros(K.int_shape(param), dtype=K.dtype(param)) for param in params]
            
            # reset the accumulated gradient every first iteration
            agrads = [K.switch(first, grad, grad + vagrad) for grad, vagrad in zip(grads, vagrads)]

        with hooks.get_gradients(self.optimizer, [agrad / self.steps for agrad in agrads]), \
            hooks.update    (condition=last, name_scope=name_scope),    \
            hooks.update_add(condition=last, name_scope=name_scope),    \
            hooks.update_sub(condition=last, name_scope=name_scope):
            for update in self.optimizer.get_updates(loss, params):
                # get_updates() may return assignment ops or tuples (variable, gradient) representing the desired assignments
                # in the latter case we want to build the assignment op ourselves under the same hooks
                #
                # reference: https://github.com/tensorflow/tensorflow/blob/v1.14.0/tensorflow/python/keras/backend.py#L3145
                if isinstance(update, tuple):
                    with K.name_scope(name_scope):
                        update = K.update(update[0], update[1])

                self.updates.append(update)

        assert K.backend() == 'tensorflow', "Unsupported backend (" + K.backend() + ")"

        with K.name_scope(name_scope):
            self.updates.extend([K.update(vagrad, agrad) for vagrad, agrad in zip(vagrads, agrads)])

            with K.get_session().graph.control_dependencies(self.updates):
                self.updates.append(K.update_add(iterations, 1))
        
        return self.updates

    def get_gradients(self, loss, params):
        return self.optimizer.get_gradients(loss, params)

    def set_weights(self, weights):
        self.optimizer.set_weights(weights)

    def get_weights(self):
        return self.optimizer.get_weights()
    
    def get_config(self):
        # we have to support creating our optimizers from configurations in order to support being run with Horovod
        # Horovod dynamically creates a class that inherits the optimizer class it's wrapping (our optimizers), and
        # passes the dictionary returned from this very method as the kwargs for the initialization in __init__()
        # 
        # our optimizers inherit from this very class, receive 'steps' as an argument, and do not receive 'optimizer'
        # as they create the one they mimic
        #
        # therefore, we do not save self.optimizer in the returned dictionary
        
        config = self.optimizer.get_config()
        config['steps'] = self.steps
        return config

    def __getattribute__(self, name):
        # users can query the optimizer to retrieve its attributes, such as 'lr'.
        # we rely on the fact that there are no mutual attribute names between our
        # implementation and the original optimizer implementation, and we get the
        # original optimizer's attribute in case our object does not have one.
        try:
            return object.__getattribute__(self, name)
        except AttributeError:
            return self.optimizer.__getattribute__(name)

def _optimizer(optimizer):
    setattr(
        sys.modules[__name__],
        optimizer,
        type(
            optimizer, (Optimizer,),
            { '__init__': lambda self, steps, **kwargs: Optimizer.__init__(self, optimizer=getattr(keras.optimizers, optimizer)(**kwargs), steps=steps) }
        )
    )

[_optimizer(optimizer) for optimizer in ['SGD', 'RMSprop', 'Adagrad', 'Adadelta', 'Adam', 'Adamax', 'Nadam']]



class Hook(object):
    def __init__(self, module, method, hook=None, recursion=False):
        self.module = module
        self.method = method
        self.hook = hook if hook else self.__hook__
        self.recursion = recursion
        self.__original__ = getattr(module, method)

    def enable(self):
        def hook(*args, **kwargs):
            if hook.called and not self.recursion:
                return self.__original__(*args, **kwargs)
            
            hook.called = True
            result = self.hook(*args, **kwargs)
            hook.called = False
            return result
        
        hook.called = False
        
        setattr(self.module, self.method, hook)

    def disable(self):
        setattr(self.module, self.method, self.__original__)

    def __enter__(self):
        self.enable()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.disable()
    
    def __hook__(self):
        raise NotImplementedError()



from __future__ import absolute_import # to support 'import random' in Python 2

import random as r
import string as s

number = lambda a=10, b=100: r.randint(a, b)

def string(length=5):
    return ''.join(r.choice(s.ascii_letters + s.digits + s.punctuation) for _ in range(length))

def strings(count=number(2, 10)):
    return [string() for _ in range(count)]

def shape():
    return tuple([number(2, 5) for _ in range(number(2, 4))])



from __future__ import absolute_import # to support 'import keras.backend' in Python 2

import runai.elastic
import runai.utils

from . import models

def init(global_batch_size, max_gpu_batch_size, gpus=None):
    # first of all calculate the number of GA steps and the batch size
    runai.elastic._init(global_batch_size, max_gpu_batch_size, gpus)

    # now use Horovod if needed
    if runai.elastic.gpus > 1:
        runai.utils.log.debug('Initializing Horovod')
        import horovod.keras as hvd
        hvd.init()

        setattr(runai.elastic, 'master', hvd.local_rank() == 0)
        setattr(runai.elastic, 'hvd', hvd) # so that anyone will be easily accessible to Horovod

        runai.utils.log.debug('Attaching Keras session to GPU #%d', hvd.local_rank())
        import tensorflow
        config = tensorflow.ConfigProto()
        config.gpu_options.visible_device_list = str(hvd.local_rank())
        import keras.backend
        keras.backend.set_session(tensorflow.Session(config=config)) # TODO(levosos): support cases where configuration will be set afterwards



def rename(obj, old, new):
    setattr(obj, new, getattr(obj, old))
    delattr(obj, old)

class Attribute(object):
    """ A utility class for creating temporary attributes in objects
    """

    def __init__(self, obj, name, value):
        assert isinstance(name, list) == isinstance(value, list)

        self.obj = obj
        self.name = name
        self.value = value

    def __enter__(self):
        if isinstance(self.name, list):
            for name, value in zip(self.name, self.value):
                setattr(self.obj, name, value)
        else:
            setattr(self.obj, self.name, self.value)

        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if isinstance(self.name, list):
            for name in self.name:
                delattr(self.obj, name)
        else:
            delattr(self.obj, self.name)



import logging
import sys

FORMAT      = '[Run:AI] [%(levelname)-8s] [%(asctime)s.%(msecs)03d] [%(process)d] [%(filename)-23s:%(lineno)-4d] %(message)s'
DATE_FMT    = '%d-%m-%Y %H:%M:%S'

# FORMATTER
try:
    import coloredlogs
    formatter = coloredlogs.ColoredFormatter(
        fmt             = FORMAT,
        datefmt         = DATE_FMT,
        level_styles    = dict(
            debug       = dict(
                color   = 'blue',
            ),
            info        = dict(
                color   = 'green',
            ),
            warning     = dict(
                color   = 'yellow',
            ),
            error       = dict(
                color   = 'red',
            ),
            critical    = dict(
                color       = 'white',
                background  = 'red',
            ),
        ),
        field_styles = dict(),
    )
except ImportError:
    formatter = logging.Formatter(
        fmt     = FORMAT,
        datefmt = DATE_FMT,
    )

# LOG
log = logging.getLogger("runai")
log.setLevel(logging.DEBUG)

# STDERR
stderr_handler = logging.StreamHandler(stream=sys.stderr)
stderr_handler.setFormatter(formatter)
log.addHandler(stderr_handler)

debug       = log.debug
info        = log.info
warning     = log.warning
error       = log.error
critical    = log.critical



import runai.elastic

def init(global_batch_size, max_gpu_batch_size, gpus=None):
    if gpus is None:
        import torch.cuda
        gpus = torch.cuda.device_count()

    runai.elastic._init(global_batch_size, max_gpu_batch_size, gpus)



from __future__ import absolute_import # to support 'import keras.backend' in Python 2

import enum

class Method(enum.Enum):
    Cin  = 0
    Cout = 1

from . import keras

import runai.utils

def init(splits, method):
    if method != Method.Cin and method != Method.Cout:
        raise ValueError('Unrecognized method: %s' % method)

    runai.utils.log.info('Initializing MP (%s) with %s splits', method.name, splits)
    
    import sys

    setattr(sys.modules[__name__], 'splits', splits)
    setattr(sys.modules[__name__], 'method', method)

    import keras.layers

    from .keras import layers

    [setattr(keras.layers, attribute, getattr(layers, attribute)) for attribute in [
        'Activation',
        'add',
        'Add',
        'average',
        'Average',
        'AveragePooling2D',
        'BatchNormalization',
        'Conv2D',
        'Dense',
        'Dropout',
        'Flatten',
        'GlobalAveragePooling2D',
        'GlobalMaxPooling2D',
        'maximum',
        'Maximum',
        'MaxPooling2D',
        'minimum',
        'Minimum',
        'multiply',
        'Multiply',
        'subtract',
        'Subtract',
        'ZeroPadding2D',
        ]]



import sys

import runai.utils

# users should not call this method directly, but call the `init()` method of
# the sub-package of the respecive framework (e.g. `runai.elastic.keras.init()`)
def _init(global_batch_size, max_gpu_batch_size, gpus=None):
    if gpus is None:
        gpus = runai.utils.gpus.count()

    if gpus < 1:
        raise ValueError('GPU count (%d) must be at least 1' % gpus)

    module = sys.modules[__name__]

    setattr(module, 'global_batch_size', global_batch_size)
    setattr(module, 'gpus', gpus)
    setattr(module, 'master', True)

    # TODO(levosos): support uneven dividing
    steps = max(1, global_batch_size // (max_gpu_batch_size * gpus)) # must be at least 1
    batch_size = global_batch_size // (steps * gpus)

    setattr(module, 'steps', steps)
    setattr(module, 'batch_size', batch_size)

    runai.utils.log.info('Spreading global batch size %d across %d GPU(s) each with %d step(s) of batch size %d',
        global_batch_size, gpus, steps, batch_size)



# NOTE(levosos): we do not use TensorFlow's API (tensorflow.python.client.device_lib.list_local_devices())
# because it causes all devices to be mapped to the current process and when using Horovod
# we receive the next error: "TensorFlow device (GPU:0) is being mapped to multiple CUDA devices".

import subprocess

def available():
    return count() > 0

def count():
    try:
        return int(subprocess.check_output('nvidia-smi --list-gpus | wc -l', shell=True))
    except subprocess.CalledProcessError:
        return 0



import tensorflow

from runai.utils import Hook

class make_callable(Hook):
    def __init__(self):
        super(make_callable, self).__init__(tensorflow.Session, '_make_callable_from_options')
    
    def __hook__(self, _self, callable_options):
        assert not callable_options.HasField('run_options') # TODO(levosos): support this

        # TODO(levosos): should we create a copy of 'callable_options'?
        callable_options.run_options.trace_level = tensorflow.RunOptions.FULL_TRACE

        return self.__original__(_self, callable_options)



from runai.profiler import Profiler

import tensorflow

class session_run(Profiler):
    def __init__(self, steps, dst):
        super(session_run, self).__init__(tensorflow.Session, 'run', steps, dst)
    
    def __hook__(self, _self, fetches, feed_dict=None, options=None, run_metadata=None):
        if options is None:
            options = tensorflow.RunOptions()
    
        options.trace_level = tensorflow.RunOptions.FULL_TRACE

        if run_metadata is None:
            run_metadata = tensorflow.RunMetadata()

        result = self.__original__(_self, fetches, feed_dict, options, run_metadata)

        self._update(run_metadata)

        return result



from runai.profiler import Profiler

import tensorflow
from tensorflow.python import pywrap_tensorflow as tf_c_api
from tensorflow.python.util import compat

class run_callable(Profiler):
    def __init__(self, steps, dst):
        super(run_callable, self).__init__(tf_c_api, 'TF_SessionRunCallable', steps, dst)
    
    def __hook__(self, session, handle, args, status, run_metadata_ptr):
        assert run_metadata_ptr is None # TODO(levosos): handle this

        run_metadata_ptr = tf_c_api.TF_NewBuffer()
        result = self.__original__(session, handle, args, status, run_metadata_ptr)

        proto = tf_c_api.TF_GetBuffer(run_metadata_ptr)
        tf_c_api.TF_DeleteBuffer(run_metadata_ptr)
        
        run_metadata = tensorflow.RunMetadata()
        run_metadata.ParseFromString(compat.as_bytes(proto))

        self._update(run_metadata)

        return result



import keras.backend as K
import keras.layers

import runai.mp

from .keep import Keep
from .parallelised import Parallelised

Activation = Keep.create('Activation')

class Dense(Parallelised, keras.layers.Dense):
    def build(self, input_shape):
        assert len(input_shape) == 2 # TODO(levosos): support more than two dimensions

        total_cin = input_shape[-1]

        cin, cout, c = self.calculate_cs(
            cin=total_cin,
            cout=self.units)

        self.kernels = self.add_weights(
            name='kernel',
            shape=(cin, cout),
            initializer=self.kernel_initializer,
            regularizer=self.kernel_regularizer,
            constraint=self.kernel_constraint) # TODO(levosos): is this ok?

        if self.use_bias:
            self.biases = self.add_weights(
                name='bias',
                shape=(c,),
                initializer=self.bias_initializer,
                regularizer=self.bias_regularizer,
                constraint=self.bias_constraint) # TODO(levosos): is this ok?
        
        self.input_spec = keras.layers.InputSpec(ndim=2, axes={-1: total_cin}) # TODO(levosos): use 'min_ndim' once supporting more than two dimensions
        self.built = True

    def call(self, inputs):
        inputs = self.inputs(inputs, channel_axis=-1)

        outputs = self.parallelise(
            lambda input, kernel: K.dot(input, kernel),
            inputs,
            self.kernels)

        if runai.mp.method == runai.mp.Method.Cin:
            outputs = self.reduce_split(outputs, channel_axis=-1)

        if self.use_bias:
            outputs = self.parallelise(
                lambda output, bias: K.bias_add(output, bias, data_format='channels_last'),
                outputs,
                self.biases)

        if self.activation is not None:
            outputs = self.parallelise(
                lambda output: self.activation(output),
                outputs)
        
        return self.merge(outputs, channel_axis=-1)

Dropout = Keep.create('Dropout')

Flatten = Keep.create('Flatten')



import keras.backend as K

from runai.utils import Hook

class get_gradients(Hook):
    def __init__(self, optimizer, gradients):
        super(get_gradients, self).__init__(optimizer, 'get_gradients')
        self.gradients = gradients
    
    def __hook__(self, loss, params):
        return self.gradients

class update_add(Hook):
    def __init__(self, condition, name_scope):
        super(update_add, self).__init__(K, 'update_add')
        self.condition = condition
        self.name_scope = name_scope
    
    def __hook__(self, x, increment):
        with K.name_scope(self.name_scope):
            if not K.is_tensor(increment):
                increment = K.constant(increment, dtype=K.dtype(x))

            increment = K.switch(self.condition, increment, K.constant(0, dtype=K.dtype(x)))

        return self.__original__(x, increment)

class update_sub(Hook):
    def __init__(self, condition, name_scope):
        super(update_sub, self).__init__(K, 'update_sub')
        self.condition = condition
        self.name_scope = name_scope
    
    def __hook__(self, x, decrement):
        with K.name_scope(self.name_scope):
            if not K.is_tensor(decrement):
                decrement = K.constant(decrement, dtype=K.dtype(x))

            decrement = K.switch(self.condition, decrement, K.constant(0, dtype=K.dtype(x)))

        return self.__original__(x, decrement)

class update(Hook):
    def __init__(self, condition, name_scope):
        super(update, self).__init__(K, 'update')
        self.condition = condition
        self.name_scope = name_scope
    
    def __hook__(self, x, new_x):
        with K.name_scope(self.name_scope):
            if not K.is_tensor(new_x):
                new_x = K.constant(new_x, dtype=K.dtype(x))

            new_x = K.switch(self.condition, new_x, x)

        return self.__original__(x, new_x)



import keras.layers
import tensorflow as tf

import runai.mp
import runai.utils

from . import coordinator
from .parallelised import Parallelised

ATTRIBUTES = ['gamma', 'beta', 'moving_mean', 'moving_variance']

class BatchNormalization(Parallelised, keras.layers.BatchNormalization):
    """ A MP-supported implementation for keras.layers.BatchNormalization

    The implementation is highly coupled with the Keras version, and was
    implemented in regards of Keras 2.2.4.

    We rely on the fact that specified attributes are declared in build(),
    and are used only by call().

    Instead of implementing (copying Keras's code) the layer ourselves,
    we use the original implementation and cause it for the parallelisation.
    """
    def build(self, input_shape):
        def add_weight(name, shape, *args, **kwargs):
            """ This implementation will be called every time the original
            build() method will create a weight variable. Instead, we make
            it create multiple variables with smaller channel size and placed
            on different GPUs
            """
            dim = shape[0] // runai.mp.splits # all weights are of shape (c,)
            return self.add_weights(name, (dim,), *args, **kwargs)
        
        with runai.utils.Hook(self, 'add_weight', add_weight, recursion=False): # must not allow recursion as we call add_weight() ourselves
            super(BatchNormalization, self).build(input_shape)
        
        # every attribute declared by build() is actually a list of tensors,
        # and not a single one. we make sure they are not accessed by anyone
        # by renaming them
        for attribute in ATTRIBUTES:
            runai.utils.attribute.rename(self, attribute, '_' + attribute)
    
    def call(self, inputs, *args, **kwargs):
        assert not isinstance(inputs, (tuple, list))
        
        # we expect all inputs to have underlying parallelised tensors
        # we do not support the other cases as we have already built the
        # weights for the parallelised version
        assert coordinator.registered(inputs)

        runai.utils.log.info('Using parallelised input for \'%s\' layer "%s"', self.__class__.__name__, getattr(self, 'name', 'N/A'))

        def output(gpu, input):
            values = [getattr(self, '_' + attribute) for attribute in ATTRIBUTES]
            values = [None if value is None else value[gpu] for value in values]
            
            with runai.utils.Attribute(self, ATTRIBUTES, values):
                with tf.device('/device:GPU:%d' % gpu):
                    return super(BatchNormalization, self).call(input, *args, **kwargs)

        outputs = [output(gpu, input) for gpu, input in enumerate(coordinator.resolve(inputs))]

        return self.merge(outputs, channel_axis=self.axis)



import json
import os

from tensorflow.python.client import timeline

from runai.utils import Hook

class Profiler(Hook):
    def __init__(self, module, method, steps, dst):
        super(Profiler, self).__init__(module, method)
        self._timeline = None
        self._step = 0
        self._steps = steps
        self._dst = dst
    
    def _update(self, run_metadata):
        chrome_trace = json.loads(timeline.Timeline(run_metadata.step_stats).generate_chrome_trace_format())
        
        if self._timeline is None:
            self._timeline = chrome_trace
        else:
            self._timeline['traceEvents'] += [event for event in chrome_trace['traceEvents'] if 'ts' in event]

        if self._step % self._steps == 0:
            with open(os.path.join(self._dst, 'timeline_%d' % self._step), 'w') as f:
                f.write(json.dumps(self._timeline))

        self._step += 1



import keras.backend as K
import keras.layers

import runai.mp
import runai.utils

from .keep import Keep
from .parallelised import Parallelised

class Conv2D(Parallelised, keras.layers.Conv2D):
    def build(self, input_shape):
        if self.data_format == 'channels_first':
            self.channel_axis = 1
        else:
            self.channel_axis = -1

        total_cin = input_shape[self.channel_axis]
        
        if total_cin < runai.mp.splits:
            runai.utils.log.warning('Not parallelising \'%s\' layer "%s" with input shape %s', self.__class__.__name__, getattr(self, 'name', 'N/A'), input_shape)
            self._parallelised = False
            return super(Conv2D, self).build(input_shape)

        self._parallelised = True

        cin, cout, c = self.calculate_cs(
            cin=total_cin,
            cout=self.filters)

        self.kernels = self.add_weights(
            name='kernel',
            shape=self.kernel_size + (cin, cout),
            initializer=self.kernel_initializer,
            regularizer=self.kernel_regularizer,
            constraint=self.kernel_constraint) # TODO(levosos): is this ok?

        if self.use_bias:
            self.biases = self.add_weights(
                name='bias',
                shape=(c,),
                initializer=self.bias_initializer,
                regularizer=self.bias_regularizer,
                constraint=self.bias_constraint) # TODO(levosos): is this ok?

        self.input_spec = keras.layers.InputSpec(ndim=self.rank + 2, axes={self.channel_axis: total_cin})
        self.built = True

    def call(self, inputs):
        if not self._parallelised:
            return super(Conv2D, self).call(inputs)
        
        inputs = self.inputs(inputs, channel_axis=self.channel_axis)

        assert self.rank == 2 # TODO(levosos): support other convolutions

        outputs = self.parallelise(
            lambda input, kernel: K.conv2d(
                input,
                kernel,
                strides=self.strides,
                padding=self.padding,
                data_format=self.data_format,
                dilation_rate=self.dilation_rate),
            inputs,
            self.kernels)
        
        if runai.mp.method == runai.mp.Method.Cin:
            outputs = self.reduce_split(outputs, channel_axis=self.channel_axis)

        if self.use_bias:
            outputs = self.parallelise(
                lambda output, bias: K.bias_add(output, bias, data_format=self.data_format),
                outputs,
                self.biases)

        if self.activation is not None:
            outputs = self.parallelise(
                lambda output: self.activation(output),
                outputs)
        
        return self.merge(outputs, channel_axis=self.channel_axis)

ZeroPadding2D = Keep.create('ZeroPadding2D')



import tensorflow

from .profiler import Profiler
from . import keras
from . import tf

def profile(steps, dst='/tmp/'):
    profile.session_run = tf.hooks.session_run(steps, dst)
    profile.session_run.enable()

    if '_Callable' in dir(tensorflow.Session): # Keras new API
        profile.make_callable = keras.hooks.make_callable()
        profile.run_callable = keras.hooks.run_callable(steps, dst)

        profile.make_callable.enable()
        profile.run_callable.enable()



import keras.optimizers

import runai.elastic
import runai.ga.keras
import runai.utils

def compile(self, *args, **kwargs):
    if 'optimizer' in kwargs:
        assert len(args) == 0
        optimizer = kwargs['optimizer']
    else:
        assert len(args) == 1
        optimizer = args[0]

    # TODO(levosos): support cases when 'optimizer' is not an class instance but is either a string or a dictionary
    if not isinstance(optimizer, keras.optimizers.Optimizer):
        raise ValueError("'optimizer' must be a valid keras.optimizers.Optimizer")

    runai.utils.log.debug('compile() called with optimizer %s', optimizer)

    if runai.elastic.gpus > 1:
        runai.utils.log.debug('Wrapping optimizer with Horovod')

        import horovod.keras as hvd
        optimizer = hvd.DistributedOptimizer(optimizer)

    if runai.elastic.steps > 1:
        optimizer = runai.ga.keras.optimizers.Optimizer(optimizer, runai.elastic.steps)

    kwargs['optimizer'] = optimizer

    return self.__runai__['compile'](**kwargs) # ignore 'args' as 'optimizer' is the only possible argument and it is in 'kwargs'

def fit(self, *args, **kwargs):
    runai.utils.log.debug('fit() called')

    if runai.elastic.gpus > 1:
        import horovod.keras as hvd
        callbacks = [hvd.callbacks.BroadcastGlobalVariablesCallback(0)]

        if 'callbacks' in kwargs:
            callbacks.extend(kwargs['callbacks'])

        kwargs['callbacks'] = callbacks

    return self.__runai__['fit'](*args, **kwargs)

def fit_generator(self, *args, **kwargs):
    runai.utils.log.debug('fit_generator() called')

    if runai.elastic.gpus > 1:
        import horovod.keras as hvd
        callbacks = [hvd.callbacks.BroadcastGlobalVariablesCallback(0)]

        if 'callbacks' in kwargs:
            callbacks.extend(kwargs['callbacks'])

        kwargs['callbacks'] = callbacks

    return self.__runai__['fit_generator'](*args, **kwargs)

def Model(model):
    __runai__ = {}

    for method in [compile, fit, fit_generator]: # TODO(levosos): what about evaluate() and predict()?
        __runai__[method.__name__] = getattr(model, method.__name__)
        setattr(model, method.__name__, method.__get__(model))

    setattr(model, '__runai__', __runai__)

    return model



import sys

from .keep import Keep

[setattr(sys.modules[__name__], layer, Keep.create(layer)) for layer in [
    'MaxPooling2D',
    'AveragePooling2D',
    'GlobalAveragePooling2D',
    'GlobalMaxPooling2D',
]]



import keras.layers
import tensorflow as tf

import runai.mp
import runai.utils

from . import coordinator

class Parallelised(keras.layers.Layer):
    """ A helper class for MP-supported Keras layers
    """

    def add_weights(self, name, shape, **kwargs):
        """ Declare parallelised weights

        # Arguments:
            name: The base name of the weights
            shape: The (split) shape of each weight
        
        # Returns:
            A list of tensors with the tensor names of the weights
        """

        runai.utils.log.debug('Declaring %d weights (%s) of shape %s for \'%s\' layer "%s"' % (runai.mp.splits, name, shape, self.__class__.__name__, getattr(self, 'name', 'N/A')))

        def add_weight(gpu):
            with tf.device('/device:GPU:%d' % gpu):
                return self.add_weight(
                    name='%s_%d' % (name, gpu),
                    shape=shape,
                    **kwargs)
        
        return [add_weight(gpu) for gpu in range(runai.mp.splits)]

    def calculate_cs(self, cin, cout):
        """ Calculate the 'C' dimensions

        Using the total (i.e. original) cin and cout dimension sizes,
        we calculate the split sizes of those dimensions. In addition,
        we provide an extra dimension size (c) which is the split size
        of the output dimension.

        # Arguments:
            cin: the total cin size
            cout: the total cout size
        
        # Returns:
            A tuple (cin, cout, c) of the per-GPU sizes
        """

        # TODO(levosos): support uneven division
        
        c = cout // runai.mp.splits # output dimension is always split

        if runai.mp.method == runai.mp.Method.Cin:
            cin = cin // runai.mp.splits
        elif runai.mp.method == runai.mp.Method.Cout:
            cout = c
        else:
            raise ValueError('Unrecognized MP method: %s' % runai.mp.method)
        
        return (cin, cout, c)

    def inputs(self, input, channel_axis):
        """ Get the respective split (per-GPU) inputs for an input tensor

        Returns the split inputs of a merged tensor in the correct way.
        This varies between the MP methods.
        """

        assert isinstance(input, tf.Tensor) # and not a list/tuple of tensors

        if runai.mp.method == runai.mp.Method.Cin:
            if coordinator.registered(input):
                runai.utils.log.info('Using parallelised input for \'%s\' layer "%s"', self.__class__.__name__, getattr(self, 'name', 'N/A'))
                return coordinator.resolve(input)
            else:
                runai.utils.log.warning('Splitting non-parallelised input (%s) for \'%s\' layer "%s"', input.name, self.__class__.__name__, getattr(self, 'name', 'N/A'))
                return tf.split(input, runai.mp.splits, axis=channel_axis)
        elif runai.mp.method == runai.mp.Method.Cout:
            if coordinator.registered(input):
                runai.utils.log.info('Gathering parallelised input for \'%s\' layer "%s"', self.__class__.__name__, getattr(self, 'name', 'N/A'))

                def gather(gpu):
                    with tf.device('/device:GPU:%d' % gpu):
                        return keras.layers.Concatenate(axis=channel_axis)(coordinator.resolve(input))
                
                return [gather(gpu) for gpu in range(runai.mp.splits)]
            else:
                return [input] * runai.mp.splits
        else:
            raise ValueError('Unrecognized MP method: %s' % runai.mp.method)

    def merge(self, outputs, channel_axis):
        """ Merge and register the split outputs of the layer

        Merges a parallelised layer's output tensors and register
        then in 'coordinator' for later use.
        The merge is done be concatenating the tensors on the specified
        channel axis.

        # Returns:
            A tensor representing the merged outputs of the layer
        """

        merged = keras.layers.Concatenate(axis=channel_axis)(outputs)
        coordinator.register(merged, outputs)
        return merged

    def parallelise(self, l, *iterables):
        """ Evaluate a lambda on each GPU with device placement
        """
        
        def wrap(gpu, *args):
            with tf.device('/device:GPU:%d' % gpu):
                return l(*args)

        return [wrap(gpu, *args) for gpu, args in enumerate(zip(*iterables))]

    def reduce_split(self, outputs, channel_axis):
        """ Reduce-split some tensors on a specified axis
        """

        def split(gpu):
            with tf.device('/device:GPU:%d' % gpu):
                return tf.split(outputs[gpu], runai.mp.splits, axis=channel_axis)

        splits = [split(gpu) for gpu in range(runai.mp.splits)]
        
        def output(gpu):
            with tf.device('/device:GPU:%d' % gpu):
                return keras.layers.Add()([split[gpu] for split in splits])

        return [output(gpu) for gpu in range(runai.mp.splits)]
